#ifndef ENCODECODEPLUG_HH
#define ENCODECODEPLUG_HH

class QCoreApplication;
class QCommandLineParser;

int encodeCodeplug(QCommandLineParser &parser, QCoreApplication &app);

#endif // ENCODECODEPLUG_HH
