#ifndef DMR6X2UV2TEST_HH
#define DMR6X2UV2TEST_HH


#include "libdmrconfigtest.hh"

class DMR6X2UV2Test : public UnitTestBase
{
  Q_OBJECT

public:
  explicit DMR6X2UV2Test(QObject *parent = nullptr);

private slots:
};

#endif // DMR6X2UV2TEST_HH
