#ifndef WRITECODEPLUG_HH
#define WRITECODEPLUG_HH

class QCoreApplication;
class QCommandLineParser;

int writeCodeplug(QCommandLineParser &parser, QCoreApplication &app);

#endif // WRITECODEPLUG_HH
