#ifndef ENCODECALLSIGNDB_HH
#define ENCODECALLSIGNDB_HH

class QCoreApplication;
class QCommandLineParser;

int encodeCallsignDB(QCommandLineParser &parser, QCoreApplication &app);

#endif // ENCODECALLSIGNDB_HH
