#ifndef WRITECALLSIGNDB_HH
#define WRITECALLSIGNDB_HH

class QCoreApplication;
class QCommandLineParser;

int writeCallsignDB(QCommandLineParser &parser, QCoreApplication &app);

#endif // WRITECALLSIGNDB_HH
