#include "d878uv_filereader.hh"

D878UVFileReader::D878UVFileReader(Config *config, const uint8_t *data, size_t size, QString &message)
  :D868UVFileReader(config, data, size, message)
{
  // pass...
}


