#ifndef DECODECODEPLUG_HH
#define DECODECODEPLUG_HH

class QCoreApplication;
class QCommandLineParser;

int decodeCodeplug(QCommandLineParser &parser, QCoreApplication &app);

#endif // DECODECODEPLUG_HH
